/*
 * Copyright 2016 Raspberry Pi, Rene Richarz, Gordon Henderson
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/spi/spidev.h>
  
#define MAXCOUNT 10

#define LIDBITMASK      0x04
#define SCREENOFFMASK   0x02
#define PARITYMASK      0x80
#define BRIGHTNESSMASK  0x78
#define SHUTDOWNMASK    0X01

int SPISetup (void)
{
    int fd;
    int mode = 0; 
    int bpw = 8;
    int speed = 9600;

    if ((fd = open ("/dev/spidev0.1", O_RDWR)) < 0) return 0;
    if (ioctl (fd, SPI_IOC_WR_MODE, &mode) < 0) return 0;   
    if (ioctl (fd, SPI_IOC_WR_BITS_PER_WORD, &bpw) < 0) return 0;
    if (ioctl (fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed) < 0) return 0;
 
    return fd;
 }

int SPIDataRW (int fd, unsigned char *data, int len)
{
    struct spi_ioc_transfer spi;
    memset (&spi, 0, sizeof (spi));
 
    spi.tx_buf = (unsigned long) data;
    spi.rx_buf = (unsigned long) data;
    spi.len = len;
    spi.delay_usecs = 0;
    spi.speed_hz = 9600;
    spi.bits_per_word = 8;

    return ioctl (fd, SPI_IOC_MESSAGE (1), &spi);
}

int parity7 (unsigned char data)
{
    int i;
    int p = 0;
    for (i = 0; i < 7; i++) 
    {
        if (data & 1) p = !p;
        data = data >> 1;
    }
    return p;
}

int main (int argc, char *argv[])
{
    unsigned char data, new_data;
    int count, ok;
    int shutdown, brightness, parity;

    // SPI init
    int spi = SPISetup ();
    
    if (spi < 0) 
    {
      printf ("Failed to initialise SPI driver\n");
      return 1;
    }

    // read current brightness
    count = 0;
    data = 0xff;
    do 
    {
        data = 0xff;
        ok = SPIDataRW (spi, &data, 1);
        if (ok) 
        {
            parity = (data & PARITYMASK) != 0;
            brightness  = (data & BRIGHTNESSMASK) >> 3;
            ok = (parity7 (data) == parity);
        }
    }
    while ((!ok) && (count++ < MAXCOUNT));

    if (!ok)
    {
        printf ("Failed to read\n");
        return 1;
    }

    // interpret command line
    shutdown = 0;
    if (argc != 2) 
    {
        printf ("Usage : pitop shutdown | increase | decrease | <brightness value 3-10>\n");
        return 1;
    }
    
    if (strcmp (argv[1], "increase") == 0) brightness++;
    else if (strcmp (argv[1], "decrease") == 0) brightness--;
    else if (strcmp (argv[1], "shutdown") == 0) shutdown = 1;
    else if (sscanf (argv[1], "%d", &ok) == 1) brightness = ok;
    else
    {
        printf ("Usage : pitop shutdown | increase | decrease | <brightness value 3-10>\n");
        return 1;
    }
    
    if (brightness > 10) brightness = 10;
    if (brightness < 3) brightness = 3;

    // calculate data to send
    new_data = brightness << 3;
    if (parity7 (brightness)) new_data += PARITYMASK;
    if (shutdown) new_data += SHUTDOWNMASK;
    if (parity7 (new_data & 3)) new_data += LIDBITMASK;

    // send new data until accepted
    count = 0;
    do 
    {
        data = new_data;
        ok = SPIDataRW (spi, &data, 1);
        if (ok) ok = (data & BRIGHTNESSMASK) == (new_data & BRIGHTNESSMASK);
    }
    while ((!ok) && (count++ < MAXCOUNT));

    if (!ok)
    {
        printf ("Failed to write\n");
        return 1;
    }

    return 0;
}
